#!/usr/bin/env python
from common import *
from Model.model import Model
from Model.dataset import *
import time
import sys


def minimal_dist(features1, features2, i, j):
    diff = features1[0][i] - features2[0][j]
    dist = np.sqrt(np.power(diff, 2).sum())
    return dist


def prev(ind):
    return max(0, ind-1)


def dynamic_programming(emb1, emb2):
    if not type(emb1) == list:
        emb1 = [emb1]
        emb2 = [emb2]
    mat = np.ones((emb1[0].shape[0], emb2[0].shape[0])) * np.inf

    mat[:, 0] = 0

    orig = np.zeros((emb1[0].shape[0], emb2[0].shape[0], 2), dtype=int)

    for i in range(emb1[0].shape[0]):
        for j in range(mat.shape[1]):
            data_cost = minimal_dist(emb1, emb2, i, j)

            prev_costs = []
            indexes = []

            if j == 0:
                prev_costs.append(0)
                indexes.append([-1, -1])
            else:
                col = prev(j)
                for row in range(0, i+1):
                    prev_costs.append(mat[row, col])
                    indexes.append([row, col])

            cost = data_cost + prev_costs

            arg_min = np.argmin(cost)
            orig[i, j] = indexes[arg_min]
            mat[i, j] = cost[arg_min]

    return mat, orig


def done(i, j):
    if j == -1 and i == -1:
        return True
    return False


def find_path(dp_mat, dp_orig):
    a_min = np.argmin(dp_mat[:, -1])
    path = []

    i, j = a_min, dp_orig.shape[1]-1
    cost = dp_mat[i, j]
    while not done(i, j):
        path.append([i, j])
        i, j = dp_orig[i, j]

    path = path[::-1]
    return np.array(path), cost


def run_eval(model, dataset, db_loader, device):
    start = time.time()

    print('evaluating...')

    model.eval()

    with torch.no_grad():
        batch_size = dataset.batch_size
        total_perfects = 0
        total_frames_errors = 0
        num_batches = 0

        for i, video_data in enumerate(db_loader):
            visual_input = video_data['visual_input'].to(device)
            target_prediction = video_data['target_prediction'].to(device)
            audio_features = video_data['audio_features'].to(device)
            video_seq_len = video_data['video_seq_len'].to(device)
            audio_seq_len = video_data['audio_seq_len'].to(device)

            b_vid_emb, b_aud_emb = model.get_embeddings(visual_input, audio_features)

            batch_perfects = 0
            batch_frames_errors = 0
            for i_batch in range(0, batch_size):
                vid_emb, aud_emb = b_vid_emb[i_batch].cpu().numpy(), b_aud_emb[i_batch].cpu().numpy()
                vid_seq_len = int(video_seq_len[i_batch].cpu().numpy())
                # vid_emb = vid_emb[:vid_seq_len]

                dp_mat, dp_orig = dynamic_programming(vid_emb, aud_emb)
                path, cost = find_path(dp_mat, dp_orig)

                current_prediction = np.array([x[0] for x in path])
                current_target = target_prediction[i_batch].cpu().numpy()
                current_perfects = (current_prediction == current_target).sum() / len(current_prediction)
                current_frames_errors = np.abs(current_prediction - current_target).sum() / len(current_prediction)
                # print(current_perfects)
                # print(current_frames_errors)
                batch_perfects += current_perfects
                batch_frames_errors += current_frames_errors

            batch_perfects = batch_perfects / batch_size
            total_perfects += batch_perfects

            batch_frames_errors = batch_frames_errors / batch_size
            total_frames_errors += batch_frames_errors

            print('batch perfects: {}'.format(batch_perfects))
            print('batch fe: {}'.format(batch_frames_errors))

            num_batches += 1

        total_perfects = total_perfects / num_batches
        total_frames_errors = total_frames_errors / num_batches

        print('total perfects: {}'.format(total_perfects))
        print('total fe: {}'.format(total_frames_errors))

    model.train()
    end = time.time()

    print('evaluation took {}'.format(end - start))


def main(argv):
    os.chdir('..')
    config = ConfigParser("./Model/config.yaml")
    db_config = ConfigParser("dbs_config.yaml")
    db_loader, dataset = init_dataset(config, db_config, DBType.Test, False)
    dataset_size = len(dataset)

    gpu_ids = []
    if config.general.gpu_ids != '':
        gpu_ids = np.array(config.general.gpu_ids.split(' ')).astype(np.int)
    if torch.cuda.is_available() and len(gpu_ids) != 0:
        device = torch.device('cuda:{0}'.format(gpu_ids[0]))
    else:
        device = torch.device('cpu')

    model = Model(config, syncnet_model_path='./Model/outputs/checkpoints/37_syncnet.dat')
    model.to(device)
    run_eval(model, dataset, db_loader, device)


if __name__ == '__main__':
    main(sys.argv)
