#!/usr/bin/env python
from common import *
from Model.model import Model
from Model.dataset import *
import time
import sys


def minimal_dist(features1, features2, i, j):
    diff = features1[0][i] - features2[0][j]
    dist = np.sqrt(np.power(diff, 2).sum())
    return dist


def prev(ind):
    return max(0, ind-1)


def dynamic_programming(emb1, emb2):
    if not type(emb1) == list:
        emb1 = [emb1]
        emb2 = [emb2]
    mat = np.ones((emb1[0].shape[0], emb2[0].shape[0])) * np.inf

    mat[:, 0] = 0

    orig = np.zeros((emb1[0].shape[0], emb2[0].shape[0], 2), dtype=int)

    for i in range(emb1[0].shape[0]):
        for j in range(mat.shape[1]):
            data_cost = minimal_dist(emb1, emb2, i, j)

            prev_costs = []
            indexes = []

            if j == 0:
                prev_costs.append(0)
                indexes.append([-1, -1])
            else:
                col = prev(j)
                for row in range(0, i+1):
                    prev_costs.append(mat[row, col])
                    indexes.append([row, col])

            cost = data_cost + prev_costs

            arg_min = np.argmin(cost)
            orig[i, j] = indexes[arg_min]
            mat[i, j] = cost[arg_min]

    return mat, orig


def done(i, j):
    if j == -1 and i == -1:
        return True
    return False


def find_path(dp_mat, dp_orig):
    a_min = np.argmin(dp_mat[:, -1])
    path = []

    i, j = a_min, dp_orig.shape[1]-1
    cost = dp_mat[i, j]
    while not done(i, j):
        path.append([i, j])
        i, j = dp_orig[i, j]

    path = path[::-1]
    return np.array(path), cost


def run_eval(model, dataset, db_loader, device):
    start = time.time()

    print('evaluating...')

    model.eval()

    with torch.no_grad():
        batch_size = dataset.batch_size

        assert batch_size == 1

        avg_error = 0
        counter = 0
        max_error = 0
        perfects = 0
        perfect_frames = 0
        total_frames = 0

        for i, video_data in enumerate(db_loader):
            visual_input = video_data['visual_input'].to(device)
            shift = video_data['shift'].to(device)
            audio_features = video_data['audio_features'].to(device)
            video_seq_len = video_data['video_seq_len'].to(device)
            audio_seq_len = video_data['audio_seq_len'].to(device)

            b_vid_emb, b_aud_emb = model.get_embeddings(visual_input, audio_features)

            for i_batch in range(0, batch_size):
                vid_emb, aud_emb = b_vid_emb[i_batch].cpu().numpy(), b_aud_emb[i_batch].cpu().numpy()
                vid_seq_len = int(video_seq_len[i_batch].cpu().numpy())
                #vid_emb = vid_emb[:vid_seq_len]

                dp_mat, dp_orig = dynamic_programming(vid_emb, aud_emb)
                path, cost = find_path(dp_mat, dp_orig)

                path = np.array([x[0] for x in path])
                target_shift = shift[i_batch].cpu().numpy()

                predicted = (path - np.arange(path.shape[0])).tolist()

                tmp = np.array(predicted)
                perfect_frames += len(tmp[tmp == target_shift])
                total_frames += len(tmp)

                predicted = max(predicted, key=predicted.count)

                counter += 1
                current_error = abs(predicted - target_shift)

                if current_error == 0:
                    perfects += 1

                if current_error > max_error:
                    max_error = current_error

                avg_error += current_error
                print('{0}: current error {1}, current avg {2}, shift {3}, acc {4}'.format(counter, current_error,
                                                                                           float(avg_error) / counter,
                                                                                           target_shift,
                                                                                           float(perfects) / counter))

    avg_error = float(avg_error) / counter

    print('====================')
    print('avg_error = {}'.format(avg_error))
    print('====================')
    print('max_error = {}'.format(max_error))
    print('====================')
    print('acc = {}'.format(float(perfects) / counter))
    print('====================')
    print('perfect_frames = {}'.format(float(perfect_frames) / total_frames))
    print('====================')

    print(counter)
    print(total_frames)
    model.train()
    end = time.time()

    print('evaluation took {}'.format(end - start))


def main(argv):
    os.chdir('..')
    config = ConfigParser("./Model/config.yaml")
    db_config = ConfigParser("dbs_config.yaml")
    db_loader, dataset = init_dataset(config, db_config, DBType.Test, True)
    dataset_size = len(dataset)

    gpu_ids = []
    if config.general.gpu_ids != '':
        gpu_ids = np.array(config.general.gpu_ids.split(' ')).astype(np.int)
    if torch.cuda.is_available() and len(gpu_ids) != 0:
        device = torch.device('cuda:{0}'.format(gpu_ids[0]))
    else:
        device = torch.device('cpu')

    model = Model(config, syncnet_model_path='./Model/outputs/checkpoints/37_syncnet.dat')
    model.to(device)
    run_eval(model, dataset, db_loader, device)


if __name__ == '__main__':
    main(sys.argv)
