#!/usr/bin/env python

from __future__ import print_function
import sys
from common import *
from Model.model import Model
from Model.dataset import *
import time
from collections import Counter

######################
# this script evaluates our method with the same scenario as sync net,
# to evaluate syncnet itself, run ./3rdPartiesEvals/syncnet_python/eval.py
######################

class SyncNetSingleShiftEvaluator:
    def __init__(self, db_type, config, db_config):
        super(SyncNetSingleShiftEvaluator, self).__init__()
        self.config = config
        self.db_config = db_config
        self.db_loader, self.dataset = init_dataset(self.config, self.db_config, db_type, True)
        self.dataset_size = len(self.dataset)

        self.gpu_ids = []
        if config.general.gpu_ids != '':
            self.gpu_ids = np.array(self.config.general.gpu_ids.split(' ')).astype(np.int)
        if torch.cuda.is_available() and len(self.gpu_ids) != 0:
            self.device = torch.device('cuda:{0}'.format(self.gpu_ids[0]))
        else:
            self.device = torch.device('cpu')

    def eval(self, model):
        start = time.time()

        print('evaluating...')

        model.eval()

        with torch.no_grad():
            perfects = 0
            avg_error = 0
            max_error = 0
            counter = 0
            perfect_frames = 0
            total_frames = 0

            for i, video_data in enumerate(self.db_loader):
                visual_input = video_data['visual_input'].to(self.device)

                assert visual_input.shape[0] == 1

                target_prediction = video_data['target_prediction'].to(self.device)
                audio_features = video_data['audio_features'].to(self.device)
                video_seq_len = video_data['video_seq_len'].to(self.device)
                audio_seq_len = video_data['audio_seq_len'].to(self.device)
                current_shifts = video_data['shift']
                folders_names = video_data['folder_name']
                files_names = video_data['file_name']

                _, _, _, final_outputs, _ = model(visual_input, target_prediction, audio_features, video_seq_len,
                                                  audio_seq_len, True)

                target_shift = current_shifts[0].cpu().numpy()

                predicted_shift = (torch.argmax(final_outputs, dim=-1) - torch.arange(final_outputs.shape[0], device=self.device)).tolist()

                tmp = np.array(predicted_shift)
                perfect_frames += len(tmp[tmp == target_shift])
                total_frames += len(tmp)

                predicted_shift = max(predicted_shift,key=predicted_shift.count)

                current_error = abs(predicted_shift - target_shift)

                if current_error == 0:
                    perfects += 1

                if current_error > max_error:
                    max_error = current_error

                avg_error += current_error

                counter += 1

                # print('{0}: current error {1}, current avg {2}, shift {3}, acc {4}'.format(counter, current_error,
                #                                                                            float(avg_error) / counter,
                #                                                                            target_shift,
                #                                                                            float(perfects) / counter))

            avg_error = float(avg_error) / counter

            print('====================')
            print('avg_error = {}'.format(avg_error))
            print('====================')
            print('max_error = {}'.format(max_error))
            print('====================')
            print('perfect_frames = {}'.format(float(perfect_frames) / total_frames))
            print('====================')
            print('acc = {}'.format(float(perfects) / counter))
            print('====================')

            print(counter)
            print(total_frames)

        model.train()
        end = time.time()

        print('evaluation took {}'.format(end - start))


def run_once(evaluator):
    model = Model(evaluator.config, epoch=str(37))
    # model = torch.nn.DataParallel(model)
    model.to(evaluator.device)
    evaluator.eval(model)


def run_seq(evaluator):
    for i in range(0, 59):
        print(i)
        model = Model(evaluator.config, epoch=str(i))
        # model = torch.nn.DataParallel(model)
        model.to(evaluator.device)
        evaluator.eval(model)


def main(argv):
    os.chdir('..')
    config = ConfigParser("Model/config.yaml")
    db_config = ConfigParser("dbs_config.yaml")

    db_name = db_config.general.db_name
    assert db_name == 'lrs2_dataset'

    evaluator = SyncNetSingleShiftEvaluator(DBType.Test, config, db_config)
    # multiprocessing.set_start_method('spawn', force=True)
    run_once(evaluator)
    #run_seq(evaluator)


if __name__ == '__main__':
    main(sys.argv)
